@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 from contextlib import contextmanager
-from dataclasses import dataclass, field
+from dataclasses import dataclass, field, replace
 import enum
 from functools import lru_cache
 import os
@@ -10,21 +10,34 @@ import re
 import sys
 import traceback
 from typing import (
-    Any, Iterator, List, NoReturn, Optional as Opt, TextIO, Tuple, Type,
-    TypeVar, Union
+    Any, ClassVar, Iterator, NoReturn, TYPE_CHECKING, TextIO, Type, TypeVar
 )
 
 
 liblktlang_available = True
 try:
     import liblktlang as L
-except ImportError:
+except (ImportError, OSError):
+    # ImportError may be raised if liblktlang is not available.
+    #
+    # OSError may be raised if liblktlang is available, but the underlying
+    # dynamic library is missing dependency libraries. This can occur if the
+    # library was built with one version of the compiler, and then the compiler
+    # installation gets updated.  liblktlang has stale library links and fails
+    # to load.
+    #
+    # In both scenarios, assume liblktlang is not available and continue.
     liblktlang_available = False
 
 
 from langkit.utils import Colors, assert_type, col
 
 
+if TYPE_CHECKING:
+    from langkit.compiled_types import CompiledType
+    from langkit.expressions import PropertyDef
+
+
 class DiagnosticStyle(enum.Enum):
     """Format for diagnostics that Langkit emits: location and text."""
 
@@ -59,17 +72,43 @@ class Diagnostics:
     diagnostics.
     """
 
+    blacklisted_frames: dict[str, set[int]] = {}
+    """
+    Mapping from filename to set of line numbers for all stack frames to
+    blacklist from DSL locations.
+    """
+
     @classmethod
-    def is_langkit_dsl(cls, python_file: str) -> bool:
+    def blacklist_frame(cls, frame: traceback.FrameSummary) -> None:
         """
-        Return whether ``python_file`` belongs to the user language spec.
+        Add the given frame to blacklisted ones for DSL locations.
+        """
+        if isinstance(frame.lineno, int):
+            filename = P.normpath(frame.filename)
+            lineno_set = cls.blacklisted_frames.setdefault(filename, set())
+            lineno_set.add(frame.lineno)
+
+    @classmethod
+    def is_langkit_dsl(cls, frame: traceback.FrameSummary) -> bool:
+        """
+        Return whether to exclude the given frame from locations used to create
+        diagnostics.
         """
         # If the path of the file is in the list of blacklisted paths, then
         # it's definitely not part of the language spec.
-        python_file = P.normpath(python_file)
+        python_file = P.normpath(frame.filename)
         if any(path in python_file for path in cls.blacklisted_paths):
             return False
 
+        # Never use blacklisted stack frames to create DSL locations
+        try:
+            linenos = cls.blacklisted_frames[python_file]
+        except KeyError:
+            pass
+        else:
+            if frame.lineno in linenos:
+                return False
+
         # The "manage.py" script is supposed to define settings for the
         # language spec, but is not the language spec itself.
         if "manage.py" in python_file:
@@ -122,7 +161,7 @@ class Location:
     # RA22-015 TODO: Remove this "zero if unspecified" business when we get rid
     # of the legacy DSL.
 
-    lkt_unit: Opt[L.AnalysisUnit] = field(default=None)
+    lkt_unit: L.AnalysisUnit | None = field(default=None)
 
     def gnu_style_repr(self, relative: bool = True) -> str:
         """
@@ -162,8 +201,68 @@ class Location:
         """
         return cls.from_sloc_range(node.unit, node.sloc_range)
 
+    @classmethod
+    def from_lkt_node_or_none(cls, node: L.LktNode | None) -> Location | None:
+        """
+        Create a Location based on a Lkt node. Accept null nodes: return a null
+        location in that case.
+        """
+        return None if node is None else cls.from_lkt_node(node)
 
-def extract_library_location(stack: Opt[List[Any]] = None) -> Opt[Location]:
+    @classmethod
+    def for_entity_doc(
+        cls,
+        entity: CompiledType | PropertyDef,
+    ) -> Location | None:
+        """
+        Return the location of the docstring for the given entity, if
+        available.
+        """
+        # Lkt case: we have precise location for the entity doc: just return it
+        if entity._doc_location is not None:
+            return entity._doc_location
+
+        # Python DSL: the docstring usually starts one line after the
+        # declaration's first line ("class" keyword or "def" one), and the
+        # docstring first line (just the """ string delimiter) is stripped.
+        result = entity.location
+        if result is not None:
+            result = replace(result, line=result.line + 2)
+        return result
+
+    @staticmethod
+    def resolve(location: Location | L.LktNode) -> Location:
+        """
+        If "location" is a node from liblktlang, turn it into a Location
+        instance. Note that this possible only if liblktlang is available, so
+        we know that we'll have a Location instance afterwards.
+        """
+        if liblktlang_available and isinstance(location, L.LktNode):
+            return Location.from_lkt_node(location)
+        else:
+            assert isinstance(location, Location)
+            return location
+
+    builtin: ClassVar[Location]
+    """
+    Special location to designate the abstract source location where builtins
+    are defined.
+    """
+
+    nowhere: ClassVar[Location]
+    """
+    Special location to designate no location in particular. Useful for errors
+    that do not relate to a specific place in source code.
+    """
+
+
+Location.builtin = Location("<builtin>")
+Location.nowhere = Location("")
+
+
+def extract_library_location(
+    stack: list[Any] | None = None
+) -> Location | None:
     """
     Extract the location of the definition of an entity in the language
     specification from a stack trace. Use `traceback.extract_stack()` if no
@@ -172,29 +271,32 @@ def extract_library_location(stack: Opt[List[Any]] = None) -> Opt[Location]:
     stack = stack or traceback.extract_stack()
 
     # Create Location instances for each stack frame
-    locs = [Location(file=t[0], line=t[1])
+    locs = [Location(file=t.filename, line=t.lineno)
             for t in stack
-            if Diagnostics.is_langkit_dsl(t[0])]
+            if isinstance(t.lineno, int) and Diagnostics.is_langkit_dsl(t)]
 
     return locs[-1] if locs else None
 
 
-context_stack: List[Opt[Location]] = []
+context_stack: list[Location | None] = []
 
 
 @contextmanager
-def diagnostic_context(location: Opt[Location]) -> Iterator[None]:
+def diagnostic_context(location: Location | None) -> Iterator[None]:
     """
     Context manager to temporarily append a location to the context stack for
     diagnostics.
 
     :param location: The location for diagnostics.
     """
-    context_stack.append(location)
-    try:
+    if location is not None:
+        context_stack.append(location)
+        try:
+            yield
+        finally:
+            context_stack.pop()
+    else:
         yield
-    finally:
-        context_stack.pop()
 
 
 class DiagnosticError(Exception):
@@ -225,7 +327,7 @@ def format_severity(severity: Severity) -> str:
     return col(msg, Colors.BOLD + SEVERITY_COLORS[severity])
 
 
-def get_structured_context() -> List[Location]:
+def get_structured_context() -> list[Location]:
     """
     From the context global structures, return a structured context locations
     list.
@@ -233,12 +335,13 @@ def get_structured_context() -> List[Location]:
     return list(l for l in reversed(context_stack) if l)
 
 
-def get_current_location() -> Opt[Location]:
+def get_current_location() -> Location:
     ctx = get_structured_context()
-    return ctx[0] if ctx else None
+    assert ctx
+    return ctx[0]
 
 
-def get_parsable_location() -> str:
+def get_parsable_location(location: Location | L.LktNode) -> str:
     """
     Returns an error location in the common tool parsable format::
 
@@ -271,6 +374,18 @@ def error(message: str, ok_for_codegen: bool = False) -> NoReturn:
     raise AssertionError("should not happen")
 
 
+def non_blocking_error(message: str, ok_for_codegen: bool = False) -> None:
+    """
+    Shortcut around ``check_source_language``, for non-fatal errors.
+    """
+    check_source_language(
+        False,
+        message,
+        ok_for_codegen=ok_for_codegen,
+        severity=Severity.non_blocking_error,
+    )
+
+
 def check_source_language(predicate: bool,
                           message: str,
                           severity: Severity = Severity.error,
@@ -308,10 +423,11 @@ def check_source_language(predicate: bool,
             message_lines[:1] + [indent + line for line in message_lines[1:]]
         )
 
+        location = get_current_location()
         if Diagnostics.style != DiagnosticStyle.default:
-            print('{}: {}'.format(get_parsable_location(), message))
+            print('{}: {}'.format(get_parsable_location(location), message))
         else:
-            print_error(message, get_current_location(), severity)
+            print_error(message, location, severity)
 
         if severity == Severity.error and do_raise:
             raise DiagnosticError()
@@ -398,14 +514,14 @@ class WarningSet:
             w.name for w in self.enabled_warnings
         ))
 
-    def enable(self, warning: Union[WarningDescriptor, str]) -> None:
+    def enable(self, warning: WarningDescriptor | str) -> None:
         """
         Enable the given warning in this WarningSet instance.
         """
         warn = self.lookup(warning) if isinstance(warning, str) else warning
         self.enabled_warnings.add(warn)
 
-    def disable(self, warning: Union[WarningDescriptor, str]) -> None:
+    def disable(self, warning: WarningDescriptor | str) -> None:
         """
         Disable the given warning in this WarningSet instance.
         """
@@ -420,8 +536,7 @@ class WarningSet:
         other.enabled_warnings = set(self.enabled_warnings)
         return other
 
-    def with_enabled(self,
-                     warning: Union[WarningDescriptor, str]) -> WarningSet:
+    def with_enabled(self, warning: WarningDescriptor | str) -> WarningSet:
         """
         Return a copy of this WarningSet instance where `warning` is enabled.
         """
@@ -429,8 +544,7 @@ class WarningSet:
         other.enable(warning)
         return other
 
-    def with_disabled(self,
-                      warning: Union[WarningDescriptor, str]) -> WarningSet:
+    def with_disabled(self, warning: WarningDescriptor | str) -> WarningSet:
         """
         Return a copy of this WarningSet instance where `warning` is disabled.
         """
@@ -458,7 +572,7 @@ class WarningSet:
     @classmethod
     def print_list(cls,
                    out: TextIO = sys.stdout,
-                   width: Opt[int] = None) -> None:
+                   width: int | None = None) -> None:
         """
         Display the list of available warnings in `f`.
 
@@ -484,14 +598,13 @@ class WarningSet:
                   file=out)
 
 
-def check_multiple(predicates_and_messages: List[Tuple[bool, str]],
+def check_multiple(predicates_and_messages: list[tuple[bool, str]],
                    severity: Severity = Severity.error) -> None:
     """
     Helper around check_source_language, check multiple predicates at once.
 
-    :param list[(bool, str)] predicates_and_messages: List of diagnostic
-        tuples.
-    :param Severity severity: The severity of the diagnostics.
+    :param predicates_and_messages: List of diagnostic tuples.
+    :param severity: The severity of the diagnostics.
     """
     for predicate, message in predicates_and_messages:
         check_source_language(predicate, message, severity)
@@ -500,7 +613,7 @@ def check_multiple(predicates_and_messages: List[Tuple[bool, str]],
 T = TypeVar('T')
 
 
-def check_type(obj: Any, typ: Type[T], message: Opt[str] = None) -> T:
+def check_type(obj: Any, typ: Type[T], message: str | None = None) -> T:
     """
     Like utils.assert_type, but produces a client error instead.
 
@@ -525,7 +638,7 @@ def errors_checkpoint() -> None:
 
 
 @lru_cache()
-def splitted_text(unit: L.AnalysisUnit) -> List[str]:
+def splitted_text(unit: L.AnalysisUnit) -> list[str]:
     """
     Memoized function to get the splitted text of a unit. Used to not have to
     compute this every time.
@@ -541,7 +654,10 @@ def style_diagnostic_message(string: str) -> str:
     return re.sub("`.*?`", lambda m: col(m.group(), Colors.BOLD), string)
 
 
-def source_listing(highlight_sloc: Location, lines_after: int = 0) -> str:
+def source_listing(
+    highlight_sloc: Location,
+    lines_after: int = 0,
+) -> str | None:
     """
     Create a source listing for an error message, centered around a specific
     sloc, that will be highlighted/careted, as in the following example::
@@ -566,13 +682,18 @@ def source_listing(highlight_sloc: Location, lines_after: int = 0) -> str:
     start_offset = highlight_sloc.column - 1
     end_offset = highlight_sloc.end_column - 1
 
+    # If the location is at the file termination, right after a line break,
+    # there is no useful line to print.
+    if line_nb >= len(source_buffer):
+        return None
+
     # Compute the width of the column needed to print line numbers
     line_nb_width = len(str(highlight_sloc.line + lines_after))
 
     # Precompute the format string for the listing left column
     prefix_fmt = "{{: >{}}} | ".format(line_nb_width)
 
-    def append_line(line_nb: Opt[int], line: str) -> None:
+    def append_line(line_nb: int | None, line: str) -> None:
         """
         Append a line to the source listing, given a line number and a line.
 
@@ -611,7 +732,7 @@ def source_listing(highlight_sloc: Location, lines_after: int = 0) -> str:
 
 
 def print_error(message: str,
-                location: Union[Location, L.LktNode, None],
+                location: Location | L.LktNode,
                 severity: Severity = Severity.error) -> None:
     """
     Prints an error.
@@ -624,17 +745,12 @@ def print_error(message: str,
         color = Colors.RED
     error_marker = col("{}: ".format(name), color + Colors.BOLD)
 
-    if location is None:
+    location = Location.resolve(location)
+
+    if not location.file:
         print(error_marker + message)
         return
 
-    # If "location" is a node from liblktlang, turn it into a Location
-    # instance. Note that this possible only if liblktlang is available, so we
-    # know that we'll have a Location instance afterwards.
-    if liblktlang_available and isinstance(location, L.LktNode):
-        location = Location.from_lkt_node(location)
-    assert isinstance(location, Location)
-
     # Print the basic error (with colors if in tty)
     print(
         "{}: {}{}".format(
@@ -645,8 +761,10 @@ def print_error(message: str,
     )
 
     # Print the source listing
-    if location.lkt_unit is not None:
-        print(source_listing(location))
+    if location.lkt_unit is not None and location.line > 0:
+        src_lst = source_listing(location)
+        if src_lst is not None:
+            print(src_lst)
 
 
 def print_error_from_sem_result(sem_result: L.SemanticResult) -> None:
