@@ -9,17 +9,23 @@ introduction to their usage.
 from __future__ import annotations
 
 from contextlib import AbstractContextManager
+import dataclasses
 from enum import Enum
 from funcy import lsplit_by
 from itertools import count
-from typing import Dict, List, Optional, Type, cast, overload
+from typing import Type, cast, overload
 
 from langkit import names
 from langkit.compile_context import CompileCtx, get_context
 from langkit.compiled_types import (ASTNodeType, AbstractNodeData,
-                                    CompiledType, T)
-from langkit.diagnostics import (check_source_language, diagnostic_context,
-                                 extract_library_location)
+                                    CompiledType, T, TypeRepo)
+from langkit.diagnostics import (
+    Location,
+    check_source_language,
+    diagnostic_context,
+    error,
+    extract_library_location,
+)
 from langkit.expressions import (AbstractExpression, FieldAccess, PropertyDef,
                                  Self, resolve_property, unsugar)
 
@@ -28,7 +34,7 @@ from langkit.expressions import (AbstractExpression, FieldAccess, PropertyDef,
 
 def add_env(no_parent: bool = False,
             transitive_parent: bool = False,
-            names: Optional[AbstractExpression] = None) -> AddEnv:
+            names: AbstractExpression | None = None) -> AddEnv:
     """Create a new lexical environment.
 
     This action creates an environment related to this node. Using this action
@@ -44,7 +50,7 @@ def add_env(no_parent: bool = False,
         environment. If not passed or if this is an empty array, the created
         environment is not named.
     """
-    return AddEnv(no_parent, transitive_parent, names)
+    return AddEnv(no_parent, unsugar(transitive_parent), names)
 
 
 class RefKind(Enum):
@@ -71,9 +77,9 @@ class RefKind(Enum):
 def reference(nodes: AbstractExpression,
               through: PropertyDef,
               kind: RefKind = RefKind.normal,
-              dest_env: Optional[AbstractExpression] = None,
-              cond: Optional[AbstractExpression] = None,
-              category: Optional[str] = None,
+              dest_env: AbstractExpression | None = None,
+              cond: AbstractExpression | None = None,
+              category: str | None = None,
               shed_corresponding_rebindings: bool = False) -> RefEnvs:
     """
     Reference a group of lexical environments, that will be lazily yielded by
@@ -103,7 +109,7 @@ def reference(nodes: AbstractExpression,
 
 
 def add_to_env(mappings: AbstractExpression,
-               resolver: Optional[PropertyDef] = None) -> AddToEnv:
+               resolver: PropertyDef | None = None) -> AddToEnv:
     """
     Specify elements to add to the lexical environment.
 
@@ -123,9 +129,9 @@ def add_to_env(mappings: AbstractExpression,
 
 def add_to_env_kv(key: AbstractExpression,
                   value: AbstractExpression,
-                  dest_env: Optional[AbstractExpression] = None,
-                  metadata: Optional[AbstractExpression] = None,
-                  resolver: Optional[PropertyDef] = None) -> AddToEnv:
+                  dest_env: AbstractExpression | None = None,
+                  metadata: AbstractExpression | None = None,
+                  resolver: PropertyDef | None = None) -> AddToEnv:
     """
     Specify a single element to add to the lexical environment. See
     langkit.expressions.envs.new_env_assoc for more precision about the first
@@ -138,7 +144,13 @@ def add_to_env_kv(key: AbstractExpression,
     """
     from langkit.expressions import new_env_assoc
 
-    return add_to_env(new_env_assoc(key, value, dest_env, metadata), resolver)
+    result = add_to_env(
+        new_env_assoc(key, value, dest_env, metadata), resolver
+    )
+    result.kv_params = AddToEnv.KVParams(
+        key, value, dest_env, metadata, resolver
+    )
+    return result
 
 
 def handle_children() -> HandleChildren:
@@ -180,13 +192,13 @@ class EnvSpec:
         """
         self.location = extract_library_location()
 
-        self.ast_node: Optional[ASTNodeType] = None
+        self.ast_node: ASTNodeType | None = None
         """
         ASTNodeType subclass associated to this environment specification.
         Initialized when creating ASTNodeType subclasses.
         """
 
-        self.initial_env: Optional[SetInitialEnv] = None
+        self.initial_env: SetInitialEnv | None = None
         """
         The SetInitialEnv action associated to this EnvSpec, if any.
         Initialized during the parsing of actions.
@@ -194,22 +206,34 @@ class EnvSpec:
 
         # Analyze the given list of actions
         self._parse_actions(list(actions))
+        self.pre_actions: list[EnvAction]
+        self.post_actions: list[EnvAction]
+        self.actions: list[EnvAction]
 
         self.adds_env = any(isinstance(a, AddEnv) for a in self.pre_actions)
         """
         Whether this spec create an environment.
         """
 
-    def _parse_actions(self, actions: List[EnvAction]) -> None:
+        self.properties_created = False
+        """
+        When types are defined in Lkt sources, properties are created during
+        types lowering: no need to do it in the "create_properties" method.
+        """
+
+    def _parse_actions(self, actions: list[EnvAction]) -> None:
         """
         Analyze the given list of actions and extract pre/post actions, i.e.
         actions executed before and after handling children.
         """
-        def count(cls: Type[EnvAction], sequence: List[EnvAction]) -> int:
+        def filter(
+            cls: Type[EnvAction],
+            sequence: list[EnvAction],
+        ) -> list[EnvAction]:
             """
             Return the number of ``cls`` instances in ``sequence``.
             """
-            return len([a for a in sequence if isinstance(a, cls)])
+            return [a for a in sequence if isinstance(a, cls)]
 
         # If present, allow Do actions to come before SetInitialEnv
         first_actions = []
@@ -220,22 +244,24 @@ class EnvSpec:
         # After that, allow at most one call to SetInitialEnv
         self.initial_env = None
         if actions and isinstance(actions[0], SetInitialEnv):
-            check_source_language(
-                isinstance(actions[0], SetInitialEnv),
-                'The initial environment must come first after the potential'
-                ' do()'
-            )
+            with actions[0].diagnostic_context:
+                check_source_language(
+                    isinstance(actions[0], SetInitialEnv),
+                    'The initial environment must come first after the'
+                    ' potential do()'
+                )
             self.initial_env = cast(SetInitialEnv, actions.pop(0))
             first_actions.append(self.initial_env)
-        check_source_language(
-            count(SetInitialEnv, actions) == 0,
-            "set_initial_env can only be preceded by do()"
-        )
 
-        check_source_language(
-            count(AddEnv, actions) <= 1,
-            "There can be at most one call to add_env()"
-        )
+        spurious_sie = filter(SetInitialEnv, actions)
+        if spurious_sie:
+            with spurious_sie[0].diagnostic_context:
+                error("set_initial_env can only be preceded by do()")
+
+        add_envs = filter(AddEnv, actions)
+        if len(add_envs) > 1:
+            with diagnostic_context(add_envs[1].location):
+                error("There can be at most one call to add_env()")
 
         # Separate actions that must occur before and after the handling of
         # children. Get also rid of the HandleChildren delimiter action.
@@ -243,10 +269,10 @@ class EnvSpec:
                               actions)
         post = post and post[1:]
 
-        check_source_language(
-            count(AddEnv, post) == 0,
-            'add_env() must occur before processing children'
-        )
+        post_add_envs = filter(AddEnv, post)
+        if post_add_envs:
+            with diagnostic_context(post_add_envs[0].location):
+                error('add_env() must occur before processing children')
 
         self.pre_actions = first_actions + pre
         self.post_actions = post
@@ -264,22 +290,22 @@ class EnvSpec:
     def create_internal_property(self,
                                  name: str,
                                  expr: None,
-                                 type: Optional[CompiledType]) -> None: ...
+                                 type: CompiledType | None) -> None: ...
 
     @overload
     def create_internal_property(
         self,
         name: str,
         expr: AbstractExpression,
-        type: Optional[CompiledType]
+        type: CompiledType | None,
     ) -> PropertyDef: ...
 
     def create_internal_property(
         self,
         name: str,
-        expr: Optional[AbstractExpression],
-        type: Optional[CompiledType]
-    ) -> Optional[PropertyDef]:
+        expr: AbstractExpression | None,
+        type: CompiledType | None,
+    ) -> PropertyDef | None:
         """
         Create an internal property for this env spec.
 
@@ -298,7 +324,8 @@ class EnvSpec:
         p = PropertyDef(
             expr, AbstractNodeData.PREFIX_INTERNAL,
             name=names.Name.from_lower(f"{name}_{next(self.PROPERTY_COUNT)}"),
-            public=False, type=type, ignore_warn_on_node=True
+            public=False,
+            type=type,
         )
         p._indexing_name = '_{}'.format(p.original_name.lower())
         p._original_name = p._indexing_name
@@ -311,8 +338,9 @@ class EnvSpec:
         Turn the various abstract expression attributes for this env spec into
         internal properties and add them to `astnode`.
         """
-        for action in self.actions:
-            action.create_internal_properties(self)
+        if not self.properties_created:
+            for action in self.actions:
+                action.create_internal_properties(self)
 
     def register_categories(self, context: CompileCtx) -> None:
         """
@@ -374,17 +402,23 @@ class EnvSpec:
 
 class EnvAction:
 
-    resolver: Optional[PropertyDef] = None
+    resolver: PropertyDef | TypeRepo.Defer | None = None
     """
     Some env actions use resolvers, that are property that will yield a lexical
     environment from a Node. To facilitate accessing it in general, we'll set a
     class attribute to None on the base class.
-
-    :type: PropertyDef
     """
 
-    def __init__(self) -> None:
-        self.location = extract_library_location()
+    def __init__(self, location: Location | None = None) -> None:
+        self.location = location or extract_library_location()
+
+    @property
+    def diagnostic_context(self) -> AbstractContextManager[None]:
+        """
+        Diagnostic context for env actions.
+        """
+        assert self.location is not None
+        return diagnostic_context(self.location)
 
     @property
     def str_location(self) -> str:
@@ -405,7 +439,7 @@ class EnvAction:
         pass
 
     def rewrite_property_refs(self,
-                              mapping: Dict[PropertyDef, PropertyDef]) -> None:
+                              mapping: dict[PropertyDef, PropertyDef]) -> None:
         """
         Rewrite `PropertyDef` references according to `mapping`. See
         CompileCtx.lower_properties_dispatching.
@@ -416,14 +450,26 @@ class EnvAction:
 
 
 class AddEnv(EnvAction):
-    def __init__(self,
-                 no_parent: bool = False,
-                 transitive_parent: bool = False,
-                 names: Optional[AbstractExpression] = None) -> None:
-        super().__init__()
+
+    def __init__(
+        self,
+        no_parent: bool = False,
+        transitive_parent: AbstractExpression | PropertyDef | None = None,
+        names: AbstractExpression | PropertyDef | None = None,
+        location: Location | None = None,
+    ) -> None:
+        super().__init__(location)
         self.no_parent = no_parent
-        self.transitive_parent = transitive_parent
-        self.names = names
+        if isinstance(transitive_parent, PropertyDef):
+            self.transitive_parent_prop: PropertyDef | None = transitive_parent
+        else:
+            self.transitive_parent_prop = None
+            self.transitive_parent = transitive_parent or False
+        if isinstance(names, PropertyDef):
+            self.names_prop: PropertyDef | None = names
+        else:
+            self.names_prop = None
+            self.names = names
 
     def create_internal_properties(self, env_spec: EnvSpec) -> None:
         self.transitive_parent_prop = env_spec.create_internal_property(
@@ -436,14 +482,31 @@ class AddEnv(EnvAction):
 
 class AddToEnv(EnvAction):
 
-    def __init__(self,
-                 mappings: AbstractExpression,
-                 resolver: Optional[PropertyDef]) -> None:
-        super().__init__()
-        self.mappings = mappings
+    @dataclasses.dataclass
+    class KVParams:
+        """
+        Arguments for the "add_to_env_kv()" action constructor.
+        """
+        key: AbstractExpression
+        value: AbstractExpression
+        dest_env: AbstractExpression | None
+        metadata: AbstractExpression | None
+        resolver: PropertyDef | None
+
+    def __init__(
+        self,
+        mappings: AbstractExpression | PropertyDef,
+        resolver: PropertyDef | TypeRepo.Defer | None,
+        location: Location | None = None,
+    ) -> None:
+        super().__init__(location)
+        if isinstance(mappings, PropertyDef):
+            self.mappings_prop: PropertyDef = mappings
+        else:
+            self.mappings = mappings
         self.resolver = resolver
 
-        self.mappings_prop: PropertyDef
+        self.kv_params: AddToEnv.KVParams | None = None
 
     def create_internal_properties(self, env_spec: EnvSpec) -> None:
         self.mappings_prop = env_spec.create_internal_property(
@@ -451,15 +514,14 @@ class AddToEnv(EnvAction):
         )
 
     def rewrite_property_refs(self,
-                              mapping: Dict[PropertyDef, PropertyDef]) -> None:
-        self.resolver = (
-            self.resolver
-            and mapping.get(self.resolver, self.resolver)
-        )
+                              mapping: dict[PropertyDef, PropertyDef]) -> None:
+        if self.resolver is not None:
+            resolver = resolve_property(self.resolver)
+            self.resolver = mapping.get(resolver, resolver)
 
     def check(self) -> None:
         ctx = get_context()
-        self.resolver = resolve_property(self.resolver)
+        resolver = self.resolver = resolve_property(self.resolver)
         with self.mappings_prop.diagnostic_context:
             mapping_type = self.mappings_prop.type
             if mapping_type.matches(T.env_assoc):
@@ -475,23 +537,23 @@ class AddToEnv(EnvAction):
                     ' got {} instead'.format(mapping_type.dsl_name)
                 )
 
-            if self.resolver:
+            if resolver:
                 # Ask for the creation of untyped wrappers for all
                 # properties used as entity resolvers.
-                self.resolver.require_untyped_wrapper()
+                resolver.require_untyped_wrapper()
 
                 check_source_language(
-                    self.resolver.type.matches(T.entity),
+                    resolver.type.matches(T.entity),
                     'Entity resolver properties must return entities'
-                    ' (got {})'.format(self.resolver.type.dsl_name)
+                    ' (got {})'.format(resolver.type.dsl_name)
                 )
                 check_source_language(
-                    not self.resolver.dynamic_vars,
+                    not resolver.dynamic_vars,
                     'Entity resolver properties must have no dynamically'
                     ' bound variable'
                 )
                 check_source_language(
-                    not self.resolver.natural_arguments,
+                    not resolver.natural_arguments,
                     'Entity resolver properties must have no argument'
                 )
 
@@ -501,14 +563,17 @@ class RefEnvs(EnvAction):
     Couple of a property and an expression to evaluate referenced envs.
     """
 
-    def __init__(self,
-                 resolver: PropertyDef,
-                 nodes_expr: AbstractExpression,
-                 kind: RefKind = RefKind.normal,
-                 dest_env: Optional[AbstractExpression] = None,
-                 cond: Optional[AbstractExpression] = None,
-                 category: Optional[str] = None,
-                 shed_rebindings: bool = False) -> None:
+    def __init__(
+        self,
+        resolver: PropertyDef | TypeRepo.Defer,
+        nodes_expr: AbstractExpression | PropertyDef,
+        kind: RefKind = RefKind.normal,
+        dest_env: AbstractExpression | PropertyDef | None = None,
+        cond: AbstractExpression | PropertyDef | None = None,
+        category: str | None = None,
+        shed_rebindings: bool = False,
+        location: Location | None = None,
+    ) -> None:
         """
         All nodes that nodes_expr yields must belong to the same analysis unit
         as the AST node that triggers this RefEnvs. Besides, the lexical
@@ -532,22 +597,27 @@ class RefEnvs(EnvAction):
         assert resolver
         assert nodes_expr
 
-        super().__init__()
+        super().__init__(location)
 
-        self.resolver: PropertyDef = resolver
-        self.nodes_expr = nodes_expr
+        self.resolver: PropertyDef | TypeRepo.Defer = resolver
+        if isinstance(nodes_expr, PropertyDef):
+            self.nodes_property = nodes_expr
+        else:
+            self.nodes_expr = nodes_expr
         self.kind = kind
-        self.dest_env = dest_env
-        self.cond = cond
+        if isinstance(dest_env, PropertyDef):
+            self.dest_env_prop: PropertyDef | None = dest_env
+        else:
+            self.dest_env_prop = None
+            self.dest_env = dest_env
+        if isinstance(cond, PropertyDef):
+            self.cond_prop: PropertyDef | None = cond
+        else:
+            self.cond_prop = None
+            self.cond = cond
         self.category = category and names.Name.from_lower(category)
         self.shed_rebindings = shed_rebindings
 
-        self.nodes_property: Optional[PropertyDef] = None
-        """
-        This holds the property that returns a list of nodes to pass to the
-        resolver. It is None before the property is built.
-        """
-
     def create_internal_properties(self, env_spec: EnvSpec) -> None:
         """
         Create the property that returns the list of nodes to resolve into
@@ -572,29 +642,30 @@ class RefEnvs(EnvAction):
         ctx = get_context()
         ctx.has_ref_env = True
 
-        self.resolver = resolve_property(self.resolver)
-        self.resolver.require_untyped_wrapper()
+        self.resolver = resolver = resolve_property(self.resolver)
+        resolver.require_untyped_wrapper()
 
         check_source_language(
-            self.resolver.type.matches(T.LexicalEnv),
+            resolver.type.matches(T.LexicalEnv),
             'Referenced environment resolver must return a lexical'
             ' environment (not {})'.format(
-                self.resolver.type.dsl_name
+                resolver.type.dsl_name
             )
         )
         check_source_language(
-            not self.resolver.natural_arguments,
+            not resolver.natural_arguments,
             'Referenced environment resolver must take no argument'
         )
         check_source_language(
-            not self.resolver.dynamic_vars,
+            not resolver.dynamic_vars,
             'Referenced environment resolver must have no dynamically bound'
             ' variable'
         )
 
     def rewrite_property_refs(self,
-                              mapping: Dict[PropertyDef, PropertyDef]) -> None:
-        self.resolver = mapping.get(self.resolver, self.resolver)
+                              mapping: dict[PropertyDef, PropertyDef]) -> None:
+        resolver = resolve_property(self.resolver)
+        self.resolver = mapping.get(resolver, resolver)
 
 
 class HandleChildren(EnvAction):
@@ -605,9 +676,16 @@ class HandleChildren(EnvAction):
 
 
 class SetInitialEnv(EnvAction):
-    def __init__(self, env_expr: AbstractExpression):
-        super().__init__()
-        self.env_expr = env_expr
+    def __init__(
+        self,
+        env_expr: AbstractExpression | PropertyDef,
+        location: Location | None = None,
+    ):
+        super().__init__(location)
+        if isinstance(env_expr, PropertyDef):
+            self.env_prop = env_expr
+        else:
+            self.env_expr = env_expr
 
     def create_internal_properties(self, env_spec: EnvSpec) -> None:
         self.env_prop = env_spec.create_internal_property(
@@ -616,8 +694,16 @@ class SetInitialEnv(EnvAction):
 
 
 class Do(EnvAction):
-    def __init__(self, expr: AbstractExpression) -> None:
-        self.expr = expr
+    def __init__(
+        self,
+        expr: AbstractExpression | PropertyDef,
+        location: Location | None = None,
+    ) -> None:
+        super().__init__(location)
+        if isinstance(expr, PropertyDef):
+            self.do_property = expr
+        else:
+            self.expr = expr
 
     def create_internal_properties(self, spec: EnvSpec) -> None:
         self.do_property = spec.create_internal_property(
