@@ -1,14 +1,19 @@
+from __future__ import annotations
+
+import abc
+import enum
 from functools import reduce
 import funcy
 import inspect
 
 from langkit import names
-from langkit.compiled_types import T
+from langkit.compiled_types import CompiledType, T, TypeRepo
 from langkit.diagnostics import check_source_language
 from langkit.expressions.base import (
     AbstractExpression, AbstractVariable, BasicExpr, BindingScope, CallExpr,
-    ComputingExpr, LiteralExpr, PropertyDef, Self, attr_call, construct,
-    dsl_document, expr_or_null, render, sloc_info_arg, unsugar
+    ComputingExpr, LambdaArgInfo, LiteralExpr, PropertyDef, ResolvedExpression,
+    Self, SequenceExpr, attr_call, construct, dsl_document, expr_or_null,
+    render, sloc_info_arg, unsugar
 )
 
 
@@ -31,29 +36,137 @@ def or_else(lhs, rhs):
     return BinaryBooleanOperator('or', lhs, rhs)
 
 
-class BinaryBooleanOperator(AbstractExpression):
-    AND = 'and'
-    OR = 'or'
+class BinaryOpKind(enum.Enum):
+    AND = "and"
+    OR = "or"
 
-    def __init__(self, kind, lhs, rhs):
+
+class BaseBinaryOp(AbstractExpression):
+    """
+    Base class for boolean and logic binary operators.
+    """
+    def __init__(self, kind: BinaryOpKind, lhs, rhs):
         """
-        :param str kind: Kind for this binary boolean operator
-            (short-circuiting).
-        :param AbstractExpression lhs: Left operand.
-        :param AbstractExpression rhs: Right operand.
+        :param kind: Kind for this binary boolean operator.
+        :param lhs: Left operand.
+        :param rhs: Right operand.
         """
         super().__init__()
-        assert kind in (self.AND, self.OR)
         self.kind = kind
         self.lhs = lhs
         self.rhs = rhs
 
-    def construct(self):
-        """
-        Construct a resolved expression for this.
+    def __repr__(self):
+        return (
+            f"<{type(self).__name__} {self.kind.value.capitalize()}"
+            f" at {self.location_repr}>"
+        )
 
-        :rtype: IfExpr
+
+class AbstractBinaryOp(BaseBinaryOp):
+    """
+    Base class or binary operators that work exclusively on booleans or on
+    logic equations.
+    """
+    @staticmethod
+    @abc.abstractmethod
+    def operand_type() -> CompiledType:
         """
+        Static method that returns the operand/result type for this operator.
+        """
+        pass
+
+    def construct(self) -> ResolvedExpression:
+        t = self.operand_type()
+        lhs = construct(self.lhs, t)
+        rhs = construct(self.rhs, t)
+        return self.common_construct(self.kind, lhs, rhs, self)
+
+
+class BooleanBinaryOp(AbstractBinaryOp):
+    """
+    Boolean binary operator expression.
+    """
+    @staticmethod
+    def operand_type() -> CompiledType:
+        return T.Bool
+
+    @staticmethod
+    def common_construct(
+        kind: BinaryOpKind,
+        lhs: ResolvedExpression,
+        rhs: ResolvedExpression,
+        abstract_expr: AbstractExpression,
+    ) -> ResolvedExpression:
+        """
+        Common code to build a resolved expression for and/or expressions.
+        """
+        then: ResolvedExpression
+        else_then: ResolvedExpression
+        if kind == BinaryOpKind.AND:
+            then = rhs
+            else_then = LiteralExpr("False", T.Bool)
+        else:
+            then = LiteralExpr("True", T.Bool)
+            else_then = rhs
+        return If.Expr(lhs, then, else_then, abstract_expr=abstract_expr)
+
+
+class LogicBinaryOp(AbstractBinaryOp):
+    """
+    Logic binary operator expression.
+    """
+    @staticmethod
+    def operand_type() -> CompiledType:
+        return T.Equation
+
+    @staticmethod
+    def common_construct(
+        kind: BinaryOpKind,
+        lhs: ResolvedExpression,
+        rhs: ResolvedExpression,
+        abstract_expr: AbstractExpression,
+    ) -> ResolvedExpression:
+        """
+        Common code to build a resolved expression for %and/%or expressions.
+        """
+        kind_name = kind.value.capitalize()
+        return CallExpr(
+            f"{kind_name}_Pred",
+            f"Create_{kind_name}",
+            T.Equation,
+            [lhs, rhs, sloc_info_arg(abstract_expr.location)],
+            abstract_expr=abstract_expr,
+        )
+
+
+class BinaryBooleanOperator(BaseBinaryOp):
+    """
+    Binary operator expression that works equally on booleans and logic
+    equations. Note that this is specific to the Python DSL: this kind of
+    expression does not exist in Lkt: we will get rid of this class once we can
+    stop supporting the Python DSL.
+    """
+    def __init__(
+        self,
+        kind: BinaryOpKind,
+        lhs: AbstractExpression,
+        rhs: AbstractExpression,
+    ):
+        super().__init__(kind, lhs, rhs)
+        self._is_equation = None
+
+    @property
+    def is_equation(self) -> bool:
+        """
+        Whether this expression computes an equation.
+
+        Computed during the "construct" pass.
+        """
+        assert self._is_equation is not None
+        return self._is_equation
+
+    def construct(self):
         def construct_op(op):
             return construct(op, lambda t: t in (T.Bool, T.Equation),
                              "Operands of binary logic operator must be of "
@@ -68,25 +181,13 @@ class BinaryBooleanOperator(AbstractExpression):
 
         if lhs.type is T.Bool:
             # Boolean case
-            if self.kind == self.AND:
-                then = rhs
-                else_then = LiteralExpr('False', T.Bool)
-            else:
-                then = LiteralExpr('True', T.Bool)
-                else_then = rhs
-            return If.Expr(lhs, then, else_then)
+            self._is_equation = False
+            return BooleanBinaryOp.common_construct(self.kind, lhs, rhs, self)
 
         else:
             # Equation case
-            kind_name = self.kind.capitalize()
-            return CallExpr(
-                '{}_Pred'.format(kind_name), 'Create_{}'.format(kind_name),
-                T.Equation, [lhs, rhs, sloc_info_arg(self.location)],
-                abstract_expr=self
-            )
-
-    def __repr__(self):
-        return f"<{self.kind.capitalize()} at {self.location_repr}>"
+            self._is_equation = True
+            return LogicBinaryOp.common_construct(self.kind, lhs, rhs, self)
 
 
 # noinspection PyPep8Naming
@@ -115,6 +216,46 @@ def Or(*args):
     return reduce(lambda a, b: a | b, args)
 
 
+@attr_call("any_of")
+class AnyOf(AbstractExpression):
+    """
+    Return whether ``expr`` is equal to one of ``values``.
+    """
+
+    def __init__(self, expr, *values):
+        super().__init__()
+        self.expr = expr
+        self.values = values
+
+    def construct(self) -> ResolvedExpression:
+        # Lower all sub-expressions. If the prefix is a node/entity, then
+        # we only require that other operands are nodes/entities themselves.
+        # Otherwise, operand types must match the prefix's.
+        expr = construct(self.expr)
+        expected_type: TypeRepo.Defer | CompiledType
+        if expr.type.is_ast_node:
+            expected_type = T.root_node
+        elif expr.type.is_entity_type:
+            expected_type = T.entity
+        else:
+            expected_type = expr.type
+        values = [construct(v, expected_type) for v in self.values]
+        assert len(values) >= 1
+
+        # Make sure the prefix has a result variable so that equality tests do
+        # not re-evaluate it over and over.
+        expr_var = expr.create_result_var("Any_Of_Prefix")
+        result = Eq.make_expr(expr_var, values.pop())
+        while values:
+            result = If.Expr(
+                Eq.make_expr(expr_var, values.pop()),
+                LiteralExpr("True", T.Bool),
+                result,
+            )
+
+        return SequenceExpr(expr, result, abstract_expr=self)
+
+
 @attr_call("equals")
 class Eq(AbstractExpression):
     """
@@ -443,13 +584,20 @@ class Then(AbstractExpression):
             return '<Then.Expr>'
 
     @staticmethod
-    def create_from_exprs(base, then_expr, var_expr):
+    def create_from_exprs(
+        base: AbstractExpression,
+        then_expr: AbstractExpression,
+        lambda_arg_infos: list[LambdaArgInfo],
+        var_expr: AbstractVariable,
+        default_val: AbstractExpression | None = None,
+    ):
         """
         Create a Then expression without going through a lambda. Used
         internally to constructs then expressions for the underscore operator.
         """
-        ret = Then(base, None)
+        ret = Then(expr=base, then_fn=None, default_val=default_val)
         ret.then_expr = then_expr
+        ret.lambda_arg_infos = lambda_arg_infos
         ret.var_expr = var_expr
         return ret
 
@@ -468,6 +616,7 @@ class Then(AbstractExpression):
         self.default_val = default_val
         self.var_expr = self.then_expr = None
         self.underscore_then = False
+        self.lambda_arg_infos = []
 
     def do_prepare(self):
         # If this Then was created using create_from exprs, there is no lambda
@@ -501,6 +650,10 @@ class Then(AbstractExpression):
         )
         self.var_expr.set_type(expr.type)
 
+        # Now that the variable is typed, ensure that its type annotation in
+        # the lambda expression (if present) is correct.
+        LambdaArgInfo.check_list(self.lambda_arg_infos)
+
         # Create a then-expr specific scope to restrict the span of the "then"
         # variable in the debugger.
         with PropertyDef.get_scope().new_child() as then_scope:
