@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union
+from typing import Callable, TYPE_CHECKING, Union
 
 from langkit.c_api import CAPISettings
 import langkit.compiled_types as ct
@@ -43,11 +43,11 @@ class OCamlAPISettings(AbstractAPISettings):
 
     context: CompileCtx
     c_api_settings: CAPISettings
-    type_graph: Dict[TypeOrPlaceholder, List[TypeOrPlaceholder]]
+    type_graph: dict[TypeOrPlaceholder, list[TypeOrPlaceholder]]
 
     def actual_c_type(
         self,
-        typ: Union[TypeOrPlaceholder, ct.TypeRepo.Defer]
+        typ: TypeOrPlaceholder | ct.TypeRepo.Defer,
     ) -> TypeOrPlaceholder:
         """
         Return the C type used to encode ``typ`` values.
@@ -147,7 +147,7 @@ class OCamlAPISettings(AbstractAPISettings):
         """
         return '{}_fields'.format(type.kwless_raw_name.lower)
 
-    def get_field_type(self, field: ct.Field) -> List[ct.CompiledType]:
+    def get_field_type(self, field: ct.Field) -> list[ct.CompiledType]:
         """
         Return a precice type for the given Field instance. The list contains
         the possible concrete types of the given field. If there is only
@@ -175,7 +175,7 @@ class OCamlAPISettings(AbstractAPISettings):
         # For determinism, sort the list of concrete precise types
         return sorted(concrete_precise_types, key=lambda t: t.dsl_name)
 
-    def get_field_minimal_type(self, field: ct.Field) -> List[ct.CompiledType]:
+    def get_field_minimal_type(self, field: ct.Field) -> list[ct.CompiledType]:
         """
         Return a precice type for the given Field instance. While
         get_field_type returns a concrete list of types, this function returns
@@ -190,7 +190,7 @@ class OCamlAPISettings(AbstractAPISettings):
         # For determinism, sort the list of types.
         return sorted(minimal_precise_types, key=lambda t: t.dsl_name)
 
-    def get_parse_fields(self, node: ct.ASTNodeType) -> List[ct.Field]:
+    def get_parse_fields(self, node: ct.ASTNodeType) -> list[ct.Field]:
         """
         Return all the parse fields to be exposed as a field in record of node.
 
@@ -200,7 +200,7 @@ class OCamlAPISettings(AbstractAPISettings):
             lambda x: not x.is_overriding and not x.abstract and not x.null
         )
 
-    def get_properties(self, node: ct.ASTNodeType) -> List[PropertyDef]:
+    def get_properties(self, node: ct.ASTNodeType) -> list[PropertyDef]:
         """
         Return the list of all properties that should be exposed to the user.
 
@@ -232,6 +232,7 @@ class OCamlAPISettings(AbstractAPISettings):
         return dispatch_on_type(type, [
             (ct.EnumType, lambda t: self.module_name(t)),
             (T.Token, lambda t: self.module_name(t)),
+            (T.SourceLocation, lambda t: self.module_name(t)),
             (T.Symbol, lambda t: self.module_name(t)),
             (ct.ASTNodeType, lambda t: self.module_name(t.entity)),
             (ct.EntityType, lambda _: 'EntityStruct'),
@@ -244,7 +245,7 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.BigInt, lambda t: self.module_name(t)),
         ])
 
-    def node_name(self, type: Union[ct.EntityType, ct.ASTNodeType]) -> str:
+    def node_name(self, type: ct.EntityType | ct.ASTNodeType) -> str:
         """
         Returns the OCaml name for an ASTNodeType or an EntityType.
 
@@ -265,6 +266,7 @@ class OCamlAPISettings(AbstractAPISettings):
         return dispatch_on_type(type, [
             (ct.EnumType, lambda enum: enum.api_name.camel),
             (T.Token, lambda _: 'Token'),
+            (T.SourceLocation, lambda _: 'Sloc'),
             (T.Symbol, lambda _: 'Symbol'),
             (T.Character, lambda _: 'Character'),
             (T.String, lambda _: 'StringType'),
@@ -288,11 +290,13 @@ class OCamlAPISettings(AbstractAPISettings):
             (ct.ASTNodeType, lambda _: False),
             (ct.EntityType, lambda _: True),
             (T.Token, lambda _: False),
+            (T.SourceLocation, lambda _: False),
             (T.Symbol, lambda t: False),
             (T.Bool, lambda _: False),
             (T.Int, lambda _: False),
             (T.Character, lambda _: False),
             (T.String, lambda _: False),
+            (T.TextType, lambda _: False),
             (ct.ArrayType, lambda t:
                 self.wrap_requires_context(t.element_type)),
             (ct.IteratorType, lambda t:
@@ -321,11 +325,13 @@ class OCamlAPISettings(AbstractAPISettings):
             (ct.ASTNodeType, lambda _: True),
             (ct.EntityType, lambda _: False),
             (T.Token, lambda _: False),
+            (T.SourceLocation, lambda _: True),
             (T.Symbol, lambda _: False),
             (T.Bool, lambda _: True),
             (T.Int, lambda _: True),
             (T.Character, lambda _: True),
             (T.String, lambda _: False),
+            (T.TextType, lambda _: False),
             (ct.ArrayType, lambda _: False),
             (ct.IteratorType, lambda _: False),
             (ct.StructType, lambda _: False),
@@ -341,7 +347,7 @@ class OCamlAPISettings(AbstractAPISettings):
         type: ct.CompiledType,
         convert: str,
         convert_ast_node: Callable[[ct.ASTNodeType], str],
-        from_module: Optional[ct.ASTNodeType] = None
+        from_module: ct.ASTNodeType | None = None
     ) -> str:
         """
         Return the wrap/unwrap function name used to wrap/unwrap the given
@@ -382,7 +388,7 @@ class OCamlAPISettings(AbstractAPISettings):
     def wrap_function_name(
         self,
         type: ct.CompiledType,
-        from_module: Optional[ct.ASTNodeType] = None
+        from_module: ct.ASTNodeType | None = None
     ) -> str:
         def convert_ast_node(type: ct.ASTNodeType) -> str:
             return "wrap_{}".format(type.kwless_raw_name.lower)
@@ -393,7 +399,7 @@ class OCamlAPISettings(AbstractAPISettings):
     def unwrap_function_name(
         self,
         type: ct.CompiledType,
-        from_module: Optional[ct.ASTNodeType] = None
+        from_module: ct.ASTNodeType | None = None
     ) -> str:
         def convert_ast_node(type: ct.ASTNodeType) -> str:
             # We need only one unwrap function for the entire hierarchy
@@ -405,8 +411,9 @@ class OCamlAPISettings(AbstractAPISettings):
     def wrap_value(self,
                    value: str,
                    type: ct.CompiledType,
-                   context: Optional[str],
-                   check_for_null: bool = False) -> str:
+                   context: str | None,
+                   check_for_null: bool = False,
+                   dec_ref: str = "true") -> str:
         """
         Given an expression for a low-level value and the associated type,
         return an other expression that yields the corresponding high-level
@@ -418,16 +425,25 @@ class OCamlAPISettings(AbstractAPISettings):
             checks if the underlying node value is null. If it is the case, the
             expression is evaluated to None. We check null only for entity
             types.
+        :param dec_ref: This value is passed to the wrap function to signal
+            that it can safely call the dec_ref function of the wrapped type
+            after being done.
         """
 
         def from_function(typ: ct.CompiledType, value: str) -> str:
+            dec_ref_arg = (
+                '~dec_ref:{} '.format(dec_ref)
+                if typ.is_refcounted else ''
+            )
+
             context_arg = (
                 '{} '.format(context)
                 if self.wrap_requires_context(typ) else ''
             )
 
-            return "{} {}({})".format(
+            return "{} {}{}({})".format(
                 self.wrap_function_name(typ),
+                dec_ref_arg,
                 context_arg,
                 value
             )
@@ -447,7 +463,7 @@ class OCamlAPISettings(AbstractAPISettings):
     def unwrap_value(self,
                      value: str,
                      type: ct.CompiledType,
-                     context: Optional[str],
+                     context: str | None,
                      check_for_none: bool = False) -> str:
         """
         Given an expression for a low-level value and the associated type,
@@ -486,7 +502,7 @@ class OCamlAPISettings(AbstractAPISettings):
 
     def is_struct(self,
                   type: ct.CompiledType,
-                  from_module: Optional[ct.ASTNodeType] = None) -> bool:
+                  from_module: ct.ASTNodeType | None = None) -> bool:
         """
         Return true if the given type is defined as a ctypes structure without
         a view.
@@ -503,7 +519,9 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.Int, lambda _: False),
             (T.Character, lambda _: False),
             (T.String, lambda _: True),
+            (T.TextType, lambda _: True),
             (T.Token, lambda _: True),
+            (T.SourceLocation, lambda _: True),
             (T.Symbol, lambda _: False),
             (ct.EnumType, lambda _: False),
             (ct.ASTNodeType, lambda _: False),
@@ -516,7 +534,7 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.EnvRebindings, lambda _: False),
         ])
 
-    def finalize_function(self, type: ct.CompiledType) -> Optional[str]:
+    def finalize_function(self, type: ct.CompiledType) -> str | None:
         """
         Return the name of the finalization function if a value of the given
         type must be finalized. A value needs a finalization if it is
@@ -529,7 +547,7 @@ class OCamlAPISettings(AbstractAPISettings):
             function name.
         """
 
-        def dec_ref(type: ct.CompiledType) -> Optional[str]:
+        def dec_ref(type: ct.CompiledType) -> str | None:
             if type.is_refcounted:
                 return '{}.dec_ref'.format(self.struct_name(type))
             else:
@@ -540,13 +558,14 @@ class OCamlAPISettings(AbstractAPISettings):
             (ct.ASTNodeType, lambda _: None),
             (ct.EntityType, lambda _: None),
             (T.AnalysisUnit, lambda _: None),
-            (T.String, lambda _: 'StringType.dec_ref'),
+            (T.String, lambda _: 'StringType.string_dec_ref'),
+            (T.TextType, lambda _: 'Text.destroy_text'),
             (ct.CompiledType, lambda t: dec_ref(t))
         ])
 
     def c_type(self,
                type: ct.CompiledType,
-               from_module: Optional[ct.CompiledType] = None) -> str:
+               from_module: ct.CompiledType | None = None) -> str:
         """
         Return the name of the OCaml ctypes value defining the type to use in
         the C API for ``type``. For ctypes, types passed to the foreign
@@ -565,6 +584,8 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.Character, lambda t: '{}.c_type'.format(self.module_name(t))),
             (T.String, lambda t: '{}.c_type'.format(self.module_name(t))),
             (T.Token, lambda t: '{}.c_type'.format(self.module_name(t))),
+            (T.SourceLocation, lambda t:
+                '{}.c_type'.format(self.module_name(t))),
             (T.Symbol, lambda t: '{}.c_type'.format(self.module_name(t))),
             (ct.EnumType, lambda t: '{}.c_type'.format(self.module_name(t))),
             (ct.ASTNodeType, lambda _: '(ptr void)'),
@@ -581,7 +602,7 @@ class OCamlAPISettings(AbstractAPISettings):
 
     def c_value_type(self,
                      type: ct.CompiledType,
-                     from_module: Optional[ct.CompiledType] = None) -> str:
+                     from_module: ct.CompiledType | None = None) -> str:
         """
         Return the name of the type to use in the C API for ``type``. This
         is the type of a c value.
@@ -599,6 +620,8 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.String, lambda _: 'string'),
             (T.Token, lambda t:
                 '{}.t structure'.format(self.module_name(t))),
+            (T.SourceLocation, lambda t:
+                '{}.t structure'.format(self.module_name(t))),
             (T.Symbol, lambda t:
                 '{}.t structure'.format(self.module_name(t))),
             (ct.EnumType, lambda t: self.type_public_name(t)),
@@ -619,7 +642,7 @@ class OCamlAPISettings(AbstractAPISettings):
 
     def type_public_name(self,
                          type: ct.CompiledType,
-                         from_module: Optional[ct.CompiledType] = None) -> str:
+                         from_module: ct.CompiledType | None = None) -> str:
         """
         Return the public API name for a given CompiledType instance.
 
@@ -642,6 +665,7 @@ class OCamlAPISettings(AbstractAPISettings):
             (T.Bool, lambda _: 'bool'),
             (T.Int, lambda _: 'int'),
             (T.Token, lambda t: from_module_name(t)),
+            (T.SourceLocation, lambda t: from_module_name(t)),
             (T.Symbol, lambda _: 'string'),
             (ct.EnumType, lambda t: from_module_name(t)),
             (ct.ASTNodeType, lambda t: self.type_public_name(t.entity)),
@@ -678,15 +702,13 @@ class OCamlAPISettings(AbstractAPISettings):
         """
         self.add_dep(typ, typ.element_type)
 
-    def ordered_types(self) -> List[TypeOrPlaceholder]:
+    def ordered_types(self) -> list[TypeOrPlaceholder]:
         """
         Return all the types sorted so that if type T1 depends on type T2, T2
         appears before T1. Returns the topological order of the types.
-
-        :rtype: list[ct.CompiledType]
         """
         marks = {typ: 'white' for typ in self.type_graph}
-        topo: List[TypeOrPlaceholder] = []
+        topo: list[TypeOrPlaceholder] = []
 
         def dfs(vertex: TypeOrPlaceholder) -> None:
             if marks[vertex] == 'black':
